/*
   This file is part of the Dvegas library for adaptive Monte Carlo
   integration inspired by Lepage's VEGAS algorithm.
   Copyright (C) 2000-2002, 2012 Nikolas Kauer, Martin Wiebusch

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  default integration/summation range:
  (0, 1) for continuous dimensions
  {0, 1, 2, ..., dDimSizes[dim]-1} for discrete dimensions
  ----------------------------------------
  random numbers in (0, 1) are provided
  to facilitate the addition of auxilliary
  dimensions that are not VEGAS driven
*/

#ifndef DVEGAS_H
#define DVEGAS_H

#include <iostream>
using std::ostream;
using std::istream;
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <stdexcept>
using std::out_of_range;
#include <cmath>
using std::fabs;
#include <stdint.h>

class OmniParameters;     // defined in omnicomp.hh

namespace HepSource
{
// MW edit: use stdint.h types for integers
typedef int32_t Int32;
typedef int64_t Int64;
typedef double Float64;

// MW edit: make Integrand type a functor
 class Integrand {
 private:
     int _cDim, _aDim, _f;
     std::vector<int> _dDimSizes;
 public:
     // constructors
     Integrand() : _cDim(1), _aDim(0), _f(1) { }
     Integrand(const int cDim, const std::vector<int>& dDimSizes,
               const int aDim, const int f) :
         _cDim(cDim), _aDim(aDim), _f(f), _dDimSizes(dDimSizes) { }
     Integrand(const int cDim, const int aDim, const int f) :
         _cDim(cDim), _aDim(aDim), _f(f) { }
     Integrand(const int cDim, const int f) :
         _cDim(cDim), _aDim(0), _f(f) { }

     // copy constructor
     Integrand(const Integrand& i) : _cDim(i._cDim), _aDim(i._aDim), _f(i._f),
             _dDimSizes(i._dDimSizes) { }
     
     // assignment operator
     virtual Integrand& operator=(const Integrand& i) {
         _cDim = i._cDim;
         _aDim = i._aDim;
         _f    = i._f;
         _dDimSizes = i._dDimSizes;
         return *this;
     }

     virtual void operator()(const double x[], const int k[], const double& weight, const double aux[], double f[]) = 0;
     virtual void beginIteration(Int64 nshots) { }
     virtual void endIteration(Int64 nshots) { }

     int getIntegrands() const { return _f; }
     virtual void setIntegrands(int f) { _f = f; }

     int getContinuousDimensions() const { return _cDim; }
     virtual void setContinuousDimensions(int cDim) { _cDim = cDim; }

     int getAuxiliaryDimensions() const { return _aDim; }
     virtual void setAuxiliaryDimensions(int aDim) { _aDim = aDim; }

     const std::vector<int>& getDiscreteDimensions() const { return _dDimSizes; }
     virtual void setDiscreteDimensions(const std::vector<int>& dDimSizes)
     { _dDimSizes = dDimSizes; }
 };

 // MW edit: wrapper class for old integrand function type
 class IntegrandFunctor : public Integrand {
 public:
     typedef void function_type(const double x[], const int k[], const double& weight, const double aux[], double f[]);
 private:
     function_type* const _i;
 public:
     IntegrandFunctor(function_type *const i) : _i(i) { }

     virtual void operator()(const double x[], const int k[], const double& weight, const double aux[], double f[])
     { _i(x,k,weight,aux,f); }
 };

// MW edit: smart integrand class which handles correlations and dynamically
// toggles adaptation for individual axes
class SmartIntegrand : public Integrand {
private:
    int _nDim;
    std::vector<bool> _adapt;
    std::vector<std::vector<int> > _corr;
    std::vector<int> _corrDim, _cAxes, _aAxes;
    void _update();
    bool _has_axis(int axis);
    
public:
    // constructors
    SmartIntegrand() : Integrand(), _nDim(1), _adapt(1, true)
    { for(int i=0; i<_nDim; ++i) _cAxes.push_back(i); }
    SmartIntegrand(int nDim, const std::vector<int>& dDimSizes, int f) :
        Integrand(nDim, dDimSizes, 0, f), _nDim(nDim), _adapt(nDim, true)
    { for(int i=0; i<_nDim; ++i) _cAxes.push_back(i); }
    SmartIntegrand(int nDim, int f) : Integrand(nDim, std::vector<int>(), 0, f),
                                      _nDim(nDim), _adapt(nDim, true)
    { for(int i=0; i<_nDim; ++i) _cAxes.push_back(i); }

    // copy constructor
    SmartIntegrand(const SmartIntegrand& i) :
    Integrand(i), _nDim(i._nDim), _adapt(i._adapt), _corr(i._corr),
    _corrDim(i._corrDim), _cAxes(i._cAxes), _aAxes(i._aAxes) { }

    // assignment operator
    virtual SmartIntegrand& operator=(const SmartIntegrand& i) {
        Integrand::operator=(i);
        _nDim = i._nDim;
        _adapt = i._adapt;
        _corr = i._corr;
        _corrDim = i._corrDim;
        _cAxes = i._cAxes;
        _aAxes = i._aAxes;
        return *this;
    }

    // access number of (continuous or auxiliary) dimensions
    int getDimensions() const { return _nDim; }
    
    // toggle adaption
    bool getAdapt(int axis) const { return _adapt[axis]; }
    void setAdapt(int axis, bool b) {
        if(axis < 0 || axis >= _nDim) return;
        if(_has_axis(axis)) return;
        _adapt[axis] = b;
        _update();
    }

    // manage correlations
    int nCorrelationBlocks() const { return _corr.size(); }
    int newCorrelationBlock() {
        if(_corr.size() == 0 || _corr.rbegin()->size() != 0)
            _corr.push_back(std::vector<int>());
        return _corr.size()-1;
    }
    void deleteCorrelationBlock(int i) {
        if(i < 0 || i >= int(_corr.size())) return;
        _corr.erase(_corr.begin()+i);
        _update();
    }
    void addCorrelatedAxis(int axis) {
        if(axis < 0 || axis >= _nDim) return;
        if(_has_axis(axis)) return;
        _adapt[axis] = true;
        _corr.rbegin()->push_back(axis);
        _update();
    }
    const std::vector<int>& getCorrelationBlock(int i) const
    { return _corr[i]; }

    // access correlatedDimensions vector
    const std::vector<int>& getCorrelatedDimensions() const { return _corrDim; }

    // abstract calc method
    virtual void calc(const std::vector<double>& x,
                      const int k[],
                      const double& weight,
                      double f[]) = 0;

    // evaluation operator
    virtual void operator()(const double x[], const int k[],
                            const double& weight, const double aux[],
                            double f[]);
    
    // disable write access to number of continuous and auxiliary dimensions
    virtual void setContinuousDimensions(int n) { }
    virtual void setAuxiliaryDimensions(int n) { }
};

enum Sampling { NONE, IMPORTANCE, STRATIFIED };

void terminate();     // wrapper, defined in dvegas.cpp

// --------------------------------------------------------------

template<typename C>
class MultidimensionalArray
{
public:
  MultidimensionalArray(const int dim, const int length);
  void resize(const int dim, const int length);     // also yields uninitialized state
private:
  void consHelper(const int dim, const int length);
public:
  void initialize(const C& c = C());
  void reset(const C& c = C());
  MultidimensionalArray& operator=(const MultidimensionalArray& mda);
  MultidimensionalArray(const MultidimensionalArray& mda);
  virtual ~MultidimensionalArray();
  const int dim() const;
  const int length() const;
  const Int64 size() const;
  C& operator[](const Int64 sequentialIndex);
  const C& operator[](const Int64 sequentialIndex) const;
  C& operator[](const vector<int>& listOfIndexes);
  const C& operator[](const vector<int>& listOfIndexes) const;
  C& at(const Int64 sequentialIndex) throw (out_of_range);
  const C& at(const Int64 sequentialIndex) const throw (out_of_range);
  C& at(const vector<int>& listOfIndexes) throw (out_of_range);
  const C& at(const vector<int>& listOfIndexes) const throw (out_of_range);
  const vector<int> seqIdxToIndexes(const Int64 sequentialIndex) const throw (out_of_range);
  const Int64 indexesToSeqIdx(const vector<int>& listOfIndexes) const throw (out_of_range);
  const vector<C> getProjection(const int dim, const Int64 projectionIterator) const;
  void putProjection(const int dim, const Int64 projectionIterator, const vector<C>& newWeights);

  void toStream(ostream& os) const;
  void fromStream(istream& is);
  MultidimensionalArray& operator+=(const MultidimensionalArray& partial);

  typedef C value_type;
  typedef Int64 difference_type;
private:
  int _dim;
  int _length;
  Int64 _size;
  vector<C>* _v;
};

template<typename C>
inline void MultidimensionalArray<C>::initialize(const C& c) { reset(c); }     // default: c = C()

template<typename C>
inline const int MultidimensionalArray<C>::dim() const { return _dim; }

template<typename C>
inline const int MultidimensionalArray<C>::length() const { return _length; }

template<typename C>
inline const Int64 MultidimensionalArray<C>::size() const { return _size; }

template<typename C>
inline C& MultidimensionalArray<C>::operator[](const Int64 sequentialIndex) { return (*_v)[sequentialIndex]; }

template<typename C>
inline const C& MultidimensionalArray<C>::operator[](const Int64 sequentialIndex) const { return (*_v)[sequentialIndex]; }

template<typename C>
inline C& MultidimensionalArray<C>::operator[](const vector<int>& listOfIndexes) { return (*_v)[indexesToSeqIdx(listOfIndexes)]; }

template<typename C>
inline const C& MultidimensionalArray<C>::operator[](const vector<int>& listOfIndexes) const { return (*_v)[indexesToSeqIdx(listOfIndexes)]; }

template<typename C>
inline C& MultidimensionalArray<C>::at(const vector<int>& listOfIndexes) throw (out_of_range) { return (*_v)[indexesToSeqIdx(listOfIndexes)]; }

template<typename C>
inline const C& MultidimensionalArray<C>::at(const vector<int>& listOfIndexes) const throw (out_of_range) { return (*_v)[indexesToSeqIdx(listOfIndexes)]; }

// --------------------------------------------------------------

class GlobalAccumulator 
{
public:
  GlobalAccumulator(const Int64 numberOfShots);
  const Int64 numberOfShots() const;
  const double efficiency() const;     // non-zero fraction
  const double positiveContribution() const;
  const double negativeContribution() const;
  void addValue(const double value, const double weight);
    
  const Float64 sumOfIntegrandValues() const;
  const Float64 sumOfIntegrandValueSquares() const;

  void toStream(ostream& os) const;
  void fromStream(istream& is);
  GlobalAccumulator& operator+=(const GlobalAccumulator& partial);
private:
  Int64 _numberOfShots;

  Float64 _sumOfIntegrandValues;
  Float64 _sumOfIntegrandValueSquares;
  Float64 _sumOfNegativeIntegrandValues;
  Int64 _numberOfShotsWhereIntegrandIsZero;
};

inline const Int64 GlobalAccumulator::numberOfShots() const { return _numberOfShots; }

inline const double GlobalAccumulator::efficiency() const { return (_numberOfShots - _numberOfShotsWhereIntegrandIsZero)/double(_numberOfShots); }

inline const double GlobalAccumulator::positiveContribution() const { return _sumOfIntegrandValues - _sumOfNegativeIntegrandValues; }

inline const double GlobalAccumulator::negativeContribution() const { return _sumOfNegativeIntegrandValues; }

inline const Float64 GlobalAccumulator::sumOfIntegrandValues() const { return _sumOfIntegrandValues; }

inline const Float64 GlobalAccumulator::sumOfIntegrandValueSquares() const { return _sumOfIntegrandValueSquares; }
  
// --------------------------------------------------------------

class IntegrandEstimate
{
public:
  IntegrandEstimate();

  const double integral() const;
  const double standardDeviation() const;
  const double chiSquarePerIteration() const;
  const int numberOfIterations() const;

  void update(const GlobalAccumulator& acc);

  const string print() const;
  void toStream(ostream& os) const;
  void fromStream(istream& is);
private:
  double _integral;
  double _standardDeviation;
  double _chiSquarePerIteration;
  int _numberOfIterations;

  // see Eq. (2) - (6) in G.P. Lepage: VEGAS-An Adaptive Multi-dimensional
  // Integration Program, Cornell pub. CLNS-80/447, (KEK-scanned)
  Float64 _weightedSumOfEstimates;        // Sum (S_alpha/sigma^2_alpha), in Eq. (3)
  Float64 _sumOfWeights;                  // Sum (1/sigma^2_alpha), Eq. (5)
  Float64 _weightedSumOfEstimateSquares;  // Sum (S^2_alpha/sigma^2_alpha), in Eq. (6)
};

inline const double IntegrandEstimate::integral() const { return _integral; }

inline const double IntegrandEstimate::standardDeviation() const { return _standardDeviation; }

inline const double IntegrandEstimate::chiSquarePerIteration() const { return _chiSquarePerIteration; }

inline const int IntegrandEstimate::numberOfIterations() const { return _numberOfIterations; }

// --------------------------------------------------------------

class SeparabilityTestForTwoDim
{
public:
  SeparabilityTestForTwoDim(const int numberOfBinsOnEachAxis);
  void addShot(const int binIndexX, const int binIndexY, const Float64 value);
  struct SeparabilityStats
  {
    double average;
    double standardDeviation;
    SeparabilityStats(const double average, const double standardDeviation);
  };
  const SeparabilityStats separabilityStats() const;     // perfectly separable: (1, 0)
  const bool denseCoverage(const Int64 numberOfShots) const;

  void toStream(ostream& os) const;
  void fromStream(istream& is);
  SeparabilityTestForTwoDim& operator+=(const SeparabilityTestForTwoDim& partial);
private:
  int _numberOfBinsOnEachAxis;
  vector< vector<Float64> > _accumulator;
  mutable SeparabilityStats _separabilityStats;     // buffered
  mutable bool _separabilityStatsUnknown;
};

inline const bool SeparabilityTestForTwoDim::denseCoverage(const Int64 numberOfShots) const { return numberOfShots >= 100 * _numberOfBinsOnEachAxis * _numberOfBinsOnEachAxis; }

// --------------------------------------------------------------

class OmniComp;

class Dvegas
{
public:
    // MW edit: constructors for new Integrand functor
    Dvegas(const int aDim, const int f, IntegrandFunctor::function_type *const i);
    Dvegas(const int aDim, const int f, Integrand& i);

    Dvegas(const int cDim, const int cBin, const int f, IntegrandFunctor::function_type *const i,
           const Sampling sampling = IMPORTANCE);
    Dvegas(const int cDim, const int cBin, const int f, Integrand& i,
           const Sampling sampling = IMPORTANCE);

    Dvegas(const int cDim, const int cBin, const vector<int>& dDimSizes, const int f,
           IntegrandFunctor::function_type *const i, const Sampling sampling = IMPORTANCE);
    Dvegas(const int cDim, const int cBin, const vector<int>& dDimSizes, const int f,
           Integrand& i, const Sampling sampling = IMPORTANCE);

    Dvegas(const int cDim, const int cBin, const int corrDim,
           const vector<int>& dDimSizes, const int aDim, const int f, IntegrandFunctor::function_type *const i,
           const Sampling sampling = IMPORTANCE);
    Dvegas(const int cDim, const int cBin, const int corrDim,
           const vector<int>& dDimSizes, const int aDim, const int f, Integrand& i,
           const Sampling sampling = IMPORTANCE);

    Dvegas(const int cDim, const int cBin, const vector<int>& corrDim,
           const vector<int>& dDimSizes, const int aDim, const int f, IntegrandFunctor::function_type *const i,
           const Sampling sampling = IMPORTANCE);
    Dvegas(const int cDim, const int cBin, const vector<int>& corrDim,
           const vector<int>& dDimSizes, const int aDim, const int f, Integrand& i,
           const Sampling sampling = IMPORTANCE);

    // MW edit: these constructors take the dimensions from the integrand
    // functor.
    Dvegas(const int cBin, Integrand& i,
           const Sampling sampling = IMPORTANCE);
    // Dvegas(const int cBin, const int corrDim, Integrand& i,
    //        const Sampling sampling = IMPORTANCE);
    Dvegas(const int cBin, const vector<int>& corrDim, Integrand& i,
           const Sampling sampling = IMPORTANCE);
    Dvegas(const int cBin, SmartIntegrand& si, 
           const Sampling sampling = IMPORTANCE);

    Dvegas();
private:
  void initialize(const int cDim, const int cBin, const vector<int>& corrDim,
                  const vector<int>& dDimSizes, const int aDim, const int f, 
                  Integrand *const i, const Sampling sampling);
public:
  virtual ~Dvegas();
private:     // prevent copy construction and assignment
  Dvegas(const Dvegas&);
  Dvegas& operator=(const Dvegas&);

  // ***************************
public:
  class CellAccumulators;
  CellAccumulators collectData(const Int64 numberOfShots, int verbose = 1);
private:
  class AccPtrs;
  const AccPtrs evaluateIntegrand(const Int64 numberOfShots, 
    const bool delegateEvaluation);     // returned pointers own memory
  friend class OmniComp;     // performComputation() uses evaluateIntegrand()
  class CorrelationsDetector;
  class AccPtrs
  {
  public:
    vector<GlobalAccumulator> *const accumulators() const;
    CellAccumulators *const cellAccumulators() const;
    CorrelationsDetector *const correlationsDetector() const;
    AccPtrs(vector<GlobalAccumulator> *const accumulators, CellAccumulators *const
            cellAccumulators, CorrelationsDetector *const correlationsDetector);

    const string toString() const;
    void deleteAcc() const;
    AccPtrs();
    void fromString(const string& s);     // creates objects
    AccPtrs& operator+=(const AccPtrs& partial);
  private:
    vector<GlobalAccumulator>* _accumulators;
    CellAccumulators* _cellAccumulators;
    CorrelationsDetector* _correlationsDetector;
  public:     // only used in connection with return value of evaluateIntegrand()
    AccPtrs(const AccPtrs& accPtrs);
    AccPtrs& operator=(const AccPtrs& accPtrs);
  };
  struct ShotParameters
  {
    ShotParameters(const Dvegas& dvegas);
    double weight;
    vector<double> x;
    vector<int> k;
    vector<double> aux;
    vector<double> f;
    vector<int> contIndexes;
    vector<int>& discIndexes;     // discIndexes = k
  };
  // ***************************

public:
  // ***************************
  void info() const;     // print info about best estimate
  const bool isAdaptingContinuousDimensions() const;
  const double chiSquarePerIteration() const;
  // ***************************

  // ***************************
  class CellAccumulators;
  void adapt(CellAccumulators& cellAcc);     // adapt weights
  const double getCurvature() const;
  void setCurvature(const double curvature);
  const double getRootAmpl() const;
  void setRootAmpl(const double rootAmpl);
  // ***************************
  struct CellAccumulators
  {
    struct Accumulator 
    {
      Accumulator();
      Float64 sumOfIntegrandValues;
      Float64 sumOfIntegrandValueSquares;
      Int64 numberOfShots;
      void addValue(const double value);

      void toStream(ostream& os) const;
      void fromStream(istream& is);
      Accumulator& operator+=(const Accumulator& partial);
    };
    vector< vector<Accumulator> > contDimAcc;
    vector< vector<Accumulator> > discDimAcc;
    vector< MultidimensionalArray<Accumulator> > corrDimAcc;

    CellAccumulators(const Dvegas& dvegas);
    void addShot(const ShotParameters& sp);

    void toStream(ostream& os) const;
    CellAccumulators();
    void fromStream(istream& is);
    CellAccumulators& operator+=(const CellAccumulators& partial);
  };
private:
  class DiscreteWeights
  {
  public:
    DiscreteWeights(const int numberOfIndexValues);     // initialized with equal weights
    const double getIndexValueAndWeight(const double randomInZeroOne, int& indexValue) const;
    void setWeights(const vector<double>& normalizedWeights);

    void toStream(ostream& os) const;
    void fromStream(istream& is);
  private:
    vector<double> _weights;
  };
  class ContinuousWeights
  {
  public:
    ContinuousWeights(const int numberOfBins);     // initialized with equal weights
    const double getWeightAndWeightedRandomAndIndex(double& randomInZeroOne, int& binIdx) const;
    // composed of:
    const int getIndexAndNewRandom(double& randomInZeroOne) const;
    const double getWeightAndWeightedRandom(const int binIdx, double& randomInZeroOne) const;

    void setWeights(const vector<double>& normalizedBinWeights);
    const vector<double> binEdges() const;

    void toStream(ostream& os) const;
    void fromStream(istream& is);
  private:
    int _numberOfBins;
    vector<double> _binEdges;       // location of grid divisions
  };
  class CorrelatedWeights
  {
  public:
    CorrelatedWeights(const int corrDim, const int corrBin);     // not initialized
    const int dim() const;
    const int length() const;
    void initialize();     // to flat, normalized weights
    void reset();          // to flat, normalized weights
    const Int64 numberOfCells() const;
    const Float64 flatWeight() const;
    const double getIndexesAndWeight(const double randomInZeroOne, vector<int>& corrIndexes) const;
    void setWeights(const MultidimensionalArray<Float64>& normalizedWeights);
    void rebinProjection(const int dim, const vector<double>& oldBinEdges, const vector<double>& newBinEdges);
    void optimize();
    void disableOptimize();
    const bool optimizeEnabled() const;

    void toStream(ostream& os) const;
    void fromStream(istream& is);
  private:
    bool _flatGrid;
    MultidimensionalArray<Float64> _weights;
    struct Optim
    {
      bool optimizeEnabled;
      vector<Int64> idxMap;
      Float64 sumOfNonTinyWeights;
      Int64 numberOfTinyWeights;
      Float64 tinyWeight;
      void reset(const Int64 numberOfWeights);  // resizes idxMap, zeros not guaranteed
    };
    Optim _optim;
  };
public:
  void disableOptimize(const int setOfCorrelatedWeights);
  
  void reset();          // discard all data
  void resetWeights();
  void resetEstimates();

  void saveStateToFile(const string& filename = "dvegas.state") const;
  // MW edit: versions for Integrand functor added
  void restoreStateFromFile(Integrand& i, const string& filename = "dvegas.state");
  void restoreStateFromFile(IntegrandFunctor::function_type *const i, const string& filename = "dvegas.state");
  void restoreWeightsFromFile(Integrand& i, const string& filename = "dvegas.state");
  void restoreWeightsFromFile(IntegrandFunctor::function_type *const i, const string& filename = "dvegas.state");
  void saveStateToStream(ostream& os) const;
  void restoreStateFromStream(Integrand& i, istream& is);
  void restoreStateFromStream(IntegrandFunctor::function_type *const i, istream& is);
  void restoreWeightsFromStream(Integrand& i, istream& is);
  void restoreWeightsFromStream(IntegrandFunctor::function_type *const i, istream& is);
  const string saveStateToString() const;
  void restoreStateFromString(Integrand& i, const string& s);
  void restoreStateFromString(IntegrandFunctor::function_type *const i, const string& s);
  void restoreWeightsFromString(Integrand& i, const string& s);
  void restoreWeightsFromString(IntegrandFunctor::function_type *const i, const string& s);

  void printVersion() const;     // digital watermark with version information
private:
  Sampling _sampling;

  // Sizes
  int _cDim;               // number of continuous dimensions
  int _cBin;               // number of bins for continuous dimensions
  int _corrDim;            // number of correlated continuous dimensions
  vector<int> _dDimSizes;  // number of values for discrete dimensions
  int _aDim;               // number of auxilliary dimensions (no adaptation)
  int _f;                  // number of functions to integrate

  // parameters in [0,1] to control damping of weight adaptation
  double _curvature;     // 0: no damping, 1: full damping for large weights
  double _rootAmpl;      // 0: extra amplification off (fast), 1: maximal damping (no adaptation)

  Integrand* _integrand;
  bool _destroy_integrand; // MW edit: flag determining ownership of integrand functor

  // adapt()
  vector<DiscreteWeights> _discDimWeights;
  vector<ContinuousWeights> _contDimWeights;
  vector<CorrelatedWeights> _corrDimWeights;

  // info()
  vector<IntegrandEstimate> _allIterationsStats;

  // ***************************
public:
  // MW edit: allow read access to _allIterationsStats
  const IntegrandEstimate& stats(int i) const { return _allIterationsStats[i]; }
  void activateCorrelationsDetector();
private:
  bool _detectCorrelations;
  class CorrelationsDetector
  {
  public:
    CorrelationsDetector(const Dvegas& dvegas);
    void addShot(const ShotParameters& sp);
    void sort();
    const bool denseCoverage(const Int64 numberOfShots) const;
    void info() const;

    void toStream(ostream& os) const;
    CorrelationsDetector();
    void fromStream(istream& is);
    CorrelationsDetector& operator+=(const CorrelationsDetector& partial);
  private:
    struct Pair
    {
      int dim1;
      int dim2;
      SeparabilityTestForTwoDim test;
      Pair(int dim1, int dim2, SeparabilityTestForTwoDim test);

      void toStream(ostream& os) const;
      void fromStream(istream& is);
    };
    vector<Pair> _pairs;
    struct LessSeparable
    {
      bool operator()(const Pair& x, const Pair& y) const;
    };
    LessSeparable _lessSeparable;
  };
  // ***************************

  // ***************************
public:
  void attachOmniComp(OmniComp *const omniComp);
  void detachOmniComp();
private:
  OmniComp* _omniComp;
  // ***************************
};

ostream& operator<<(ostream& os, const Dvegas::CellAccumulators::Accumulator& a);
istream& operator>>(istream& is, Dvegas::CellAccumulators::Accumulator& a);

inline vector<GlobalAccumulator> *const Dvegas::AccPtrs::accumulators() const { return _accumulators; }

inline Dvegas::CellAccumulators *const Dvegas::AccPtrs::cellAccumulators() const { return _cellAccumulators; }

inline Dvegas::CorrelationsDetector *const Dvegas::AccPtrs::correlationsDetector() const { return _correlationsDetector; }

inline const bool Dvegas::isAdaptingContinuousDimensions() const { return _cDim > 0; }

inline const double Dvegas::chiSquarePerIteration() const { return _allIterationsStats[0].chiSquarePerIteration(); }

inline const double Dvegas::getCurvature() const { return _curvature; }

inline void Dvegas::setCurvature(const double curvature) { _curvature = curvature; }

inline const double Dvegas::getRootAmpl() const { return _rootAmpl; }

inline void Dvegas::setRootAmpl(const double rootAmpl) { _rootAmpl = rootAmpl; }

inline const double Dvegas::ContinuousWeights::getWeightAndWeightedRandomAndIndex(double& randomInZeroOne, int& binIdx) const
{
  binIdx = getIndexAndNewRandom(randomInZeroOne);
  return getWeightAndWeightedRandom(binIdx, randomInZeroOne);
}

inline const vector<double> Dvegas::ContinuousWeights::binEdges() const { return _binEdges; }

inline const int Dvegas::CorrelatedWeights::dim() const { return _weights.dim(); }

inline const int Dvegas::CorrelatedWeights::length() const { return _weights.length(); }

inline const Int64 Dvegas::CorrelatedWeights::numberOfCells() const { return _weights.size(); }

inline const Float64 Dvegas::CorrelatedWeights::flatWeight() const { return 1.0/numberOfCells(); }

inline void Dvegas::CorrelatedWeights::initialize() { reset(); }

inline void Dvegas::CorrelatedWeights::disableOptimize() { _optim.optimizeEnabled = false; }

inline const bool Dvegas::CorrelatedWeights::optimizeEnabled() const { return _optim.optimizeEnabled; }

inline const bool Dvegas::CorrelationsDetector::denseCoverage(const Int64 numberOfShots) const { return _pairs[0].test.denseCoverage(numberOfShots); }

inline bool Dvegas::CorrelationsDetector::LessSeparable::operator()(const Pair& x, const Pair& y) const { return fabs(1. - x.test.separabilityStats().average) > fabs(1. - y.test.separabilityStats().average); }

// Lepage's VEGAS interface
// MW edit: added verbose argument
void VEGAS(Dvegas& dvegas, const Int64 numberOfShots, const int numberOfIterations, const int init = 0, int verbose = 2);

}     // HepSource

#endif     /* DVEGAS_H */
